@@ -17,12 +17,17 @@ INSERT INTO dem.message_inbox (
 ) VALUES (
 	(select pk from dem.staff where db_user = 'any-doc'),
 	(select pk_type from dem.v_inbox_item_type where type = 'memo' and category = 'administrative'),
-	'Release Notes for GNUmed 1.1.17 (database v16.17)',
-	'GNUmed 1.1.17 Release Notes:
+	'Release Notes for GNUmed 1.1.10 (database v16.10)',
+	'GNUmed 1.1.10 Release Notes:
 
-	16.17
+	1.1.10
 
-FIX: insufficient waiting time formatting [thanks J.Busser]
+IMPROVED: arriba now at version 2.4.1
+IMPROVED: support non-blocking external apps on Windows [thanks S.Hilbert]
+IMPROVED: offline docs [thanks S.Hilbert]
+IMPROVED: russian translation [thanks YvLy]
+
+FIX: bug in calculating apparent age when patient is born later today
 ');
 
 -- --------------------------------------------------------------
